@@ -1,6 +1,6 @@
+# frozen_string_literal: false
 require 'test/unit'
 require 'tmpdir'
-require_relative 'ruby/envutil'
 
 class TestTracer < Test::Unit::TestCase
   include EnvUtil
@@ -11,7 +11,7 @@ class TestTracer < Test::Unit::TestCase
       when 1
         # do nothing
       else
-        assert_match(%r{rubygems/core_ext/kernel_require\.rb:\d+:Kernel:-:}, lines[0])
+        assert_match(%r{rubygems/core_ext/kernel_require\.rb:\d+:Kernel:<:}, lines[0])
       end
       assert_equal "#0:-e:1::-: 1", lines.last
     end
@@ -19,12 +19,7 @@ class TestTracer < Test::Unit::TestCase
 
   def test_tracer_with_option_r_without_gems
     assert_in_out_err(%w[--disable-gems -rtracer -e 1]) do |(*lines),|
-      case lines.size
-      when 1
-        # do nothing
-      else
-        flunk "unexpected output from `ruby --disable-gems -rtracer -e 1`"
-      end
+      assert_equal 1, lines.size, "unexpected output from `ruby --disable-gems -rtracer -e 1`"
       assert_equal "#0:-e:1::-: 1", lines.last
     end
   end
